/*
 * This software was developed at the National Institute of Standards and
 * Technology (NIST) by employees of the Federal Government in the course
 * of their official duties. Pursuant to title 17 Section 105 of the
 * United States Code, this software is not subject to copyright protection
 * and is in the public domain. NIST assumes no responsibility  whatsoever for
 * its use by other parties, and makes no guarantees, expressed or implied,
 * about its quality, reliability, or any other characteristic.
 */

/*
 * Defines data types and functions as described in the MBE "Concept,
 * Evaluation Plan and API".
 *
 * Version 1.
 *
 * Last Updated:  1/14/10
 */

#ifndef _MBE_API_H
#define _MBE_API_H

#include <stdint.h>

/* Data Types */

typedef struct sface
{
   uint16_t image_width,
            image_height,
            image_depth;

   uint8_t format;

   uint8_t* data;

   char* description;

   uint16_t dob,
            mob,
            yob,
            day,
            month,
            year;

   uint8_t sex,
           race;

   uint16_t weight,
            height;

} ONEFACE;


typedef struct mface
{
   unsigned int numfaces;
   ONEFACE **faces;

} MULTIFACE;


typedef struct ohos
{
   uint8_t failed;
   int16_t xleft;
   int16_t yleft;
   int16_t xright;
   int16_t yright;

} EYEPAIR;


typedef struct candidate
{
   uint8_t failed;
   uint32_t template_id;
   double similarity_score,
          probability;

} CANDIDATE;


/* Functions */

int32_t get_pid(
   char* sdk_identifier,
   char* email_addres);

int32_t get_max_template_sizes(
   uint32_t* max_enrollment_template_size,
   uint32_t* max_recognition_template_size);


#ifdef MBE_CLASS_D

int32_t initialize_frontal_pose_estimation(
   const char* configuration_location);

int32_t estimate_frontal_pose_conformance(
   const MULTIFACE* input_faces,
   double* non_frontalities);

#endif


#ifdef MBE_CLASS_A
/*
 * Verification Scenario without an enrollment database.
 */

int32_t initialize_verification(
   const char* configuration_location,
   const char** descriptions,
   const uint8_t num_descriptions);

int32_t convert_multiface_to_enrollment_template(
   const MULTIFACE* input_faces,
   uint32_t* template_size,
   uint8_t* proprietary_template);

int32_t convert_multiface_to_verification_template(
   const MULTIFACE* input_faces,
   uint32_t* template_size,
   uint8_t* proprietary_template,
   uint8_t* quality);

int32_t match_templates(
   const uint8_t* verification_template,
   const uint32_t verification_template_size,
   const uint8_t* enrollment_template,
   const uint32_t enrollment_template_size,
   double* similarity);

#endif

#if defined(MBE_CLASS_B) || defined(MBE_CLASS_C)

int32_t initialize_enrollment_session(
   const char* configuration_location,
   const char* enrollment_directory,
   const uint32_t num_persons,
   const uint32_t num_images,
   const char** descriptions,
   const uint8_t num_descriptions);

int32_t convert_multiface_to_enrollment_template(
   const MULTIFACE* input_faces,
   EYEPAIR* const* output_eyes,
   uint32_t *template_size,
   uint8_t* proprietary_template);

int32_t finalize_enrollment(
   const char* enrollment_directory,
   const char* edb_name,
   const char* edb_manifest_name);

int32_t initialize_feature_extraction_session(
   const char* configuration_location,
   const char* enrollment_directory);

int32_t convert_multiface_to_identification_template(
   const MULTIFACE* input_faces,
   EYEPAIR* const* output_eyes,
   uint32_t* template_size,
   uint8_t* identification_template);

int32_t initialize_identification_session(
   const char* config_directory,
   const char* enrollment_directory);

#endif

#ifdef MBE_CLASS_C

int32_t identify_template(
   const uint8_t* identification_template,
   const uint32_t identification_template_size,
   const uint32_t candidate_list_length,
   CANDIDATE* const * candidate_list);

#endif

#ifdef MBE_CLASS_B

int32_t verify_template(const uint8_t* verification_template,
   const uint32_t verification_template_size,
   const uint32_t enrolled_identity_claim,
   double* similarity);

#endif


#endif /* _MBE_API_H_ */

