/*
 * This software was developed at the National Institute of Standards and
 * Technology (NIST) by employees of the Federal Government in the course
 * of their official duties. Pursuant to title 17 Section 105 of the
 * United States Code, this software is not subject to copyright protection
 * and is in the public domain. NIST assumes no responsibility  whatsoever for
 * its use by other parties, and makes no guarantees, expressed or implied,
 * about its quality, reliability, or any other characteristic.
 */

/*
 * Author: George Quinn, NIST
 * Email comments to gw@nist.gov
 */

extern "C"
{
   #define MBE_CLASS_C
   #include "mbe.h"

   #define __STDC_FORMAT_MACROS
   #include <inttypes.h>			// for PRIu32
}

#include "person.h"
#include "mbe_validation.h"

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <time.h>

using namespace std;


const char* enroll_dir    = "./output/identify_1toN/templates",
          * candidate_dir = "./output/identify_1toN/candidates/",
          * enroll_file   = "./input/enroll.txt",
          * ident_file    = "./input/search.txt",
          * template_file = "./output/identify_1toN/templates.cat",
          * manifest_file = "./output/identify_1toN/templates.manifest";

const size_t num_candidates = 50;

const char* img_types[1]    = { "mugshot" };
const uint8_t num_img_types = 1;

const uint32_t num_persons   = 54,
               num_templates = 54,
               num_images    = 88;


int main(int argc, char* argv[])
{

   if (argc < 2)
      fail("usage:\n%s config_dir", argv[0]);

   const char* config_dir = argv[1];

   ifstream enroll_stream(enroll_file, ios::in | ios::binary);
   if ( !enroll_stream )
      fail("Failed to open enrollment file %s for reading.", enroll_file);


   cout << "Calling initialize_enrollment_session()." << endl;

   int32_t ret = initialize_enrollment_session(config_dir, enroll_dir,
                                               num_persons, num_images,
                                               img_types, num_img_types);

   if (ret != 0)
      fail("Bad return value from initialize_enrollment_session(): %" PRIu32 ".", ret);

 
   uint32_t max_enroll_size,
            max_ident_size;

   ret = get_max_template_sizes(&max_enroll_size, &max_ident_size);

   if (ret != 0)
      fail("Bad return value from get_max_template_sizes(): %" PRIu32 ".", ret);


   ofstream template_stream(template_file, ios::out | ios::binary);
   if ( !template_stream )
      fail("Failed to open template file %s for writing.", template_file);

   ofstream manifest_stream(manifest_file, ios::out | ios::binary);
   if ( !manifest_stream )
      fail("Failed to open manifest file %s for writing.", manifest_file);


   EYEPAIR  eye_locs[person::max_num_faces];
   EYEPAIR* eye_locs_ptrs[person::max_num_faces];

   for (int i = 0; i < person::max_num_faces; ++i)
      eye_locs_ptrs[i] = &eye_locs[i];

   uint8_t* enroll_template = new uint8_t[ max_enroll_size * person::max_num_faces ];
   uint32_t enroll_template_size;

   person p;                           // wrapper for MULTIFACE object
   uint32_t template_counter = 0;
   
   cout << "Enrolling templates." << endl;
 
   // Iterate over each person in enrollment file.
   while ( enroll_stream >> p )
   {
      ret = convert_multiface_to_enrollment_template(p.get_multiface(),
                                                     eye_locs_ptrs,
                                                     &enroll_template_size,
                                                     enroll_template);

      if (ret != 0)
      {
         cerr << "Failed to generate template for person " << p.get_pid()
              << ": " << ret << "." << endl;

      } else {

         manifest_stream << setw(3) << setfill('0') << p.get_pid() << " "
                         << enroll_template_size << " " << template_stream.tellp() << endl;
      
         // Write the template to the template concatenation file.
         template_stream.write((char*)enroll_template, enroll_template_size);

         if (template_stream.bad())
            fail("Failed to write to template concatenation file.");
      }

      ++template_counter;
   }


   if (template_counter != num_templates)
      fail("Should be %" PRIu32 " templates, but read %" PRIu32 ".", num_templates, template_counter);
 

   // House-keeping.
   enroll_stream.close();
   template_stream.close();
   manifest_stream.close();

   delete [] enroll_template;



   cout << "Calling finalize_enrollment()." << endl;

   ret = finalize_enrollment(enroll_dir, template_file, manifest_file);

   if (ret != 0)
      fail("Bad return value from finalize_enrollment: %" PRIu32, ret);
   

   cout << "Enrollment completed." << endl
        << "Calling initialize_feature_extraction_session()." << endl;

   ret = initialize_feature_extraction_session(config_dir, enroll_dir);

   if (ret != 0)
      fail("Bad return value from initialize_feature_extraction_session: %" PRIu32,
           ret);

 
   cout << "Calling initialize_identification_session()." << endl;

   ret = initialize_identification_session(config_dir, enroll_dir);

   if (ret != 0)
      fail("Bad return value from initialize_identification_session: %" PRIu32,
           ret);

   
   ifstream ident_stream(ident_file, ios::in | ios::binary);
   if ( !ident_stream )
      fail("Failed to open %s for reading.", ident_file);
   

   CANDIDATE candidates[num_candidates];
   CANDIDATE* candidate_ptrs[num_candidates];

   for (size_t i = 0; i < num_candidates; ++i)
      candidate_ptrs[i] = &candidates[i];

   uint8_t* ident_template = new uint8_t[ max_ident_size * person::max_num_faces ];
   uint32_t ident_template_size;

   cout << "Making identification templates and searching." << endl;

   // Iterate over each Person in search file.
   while ( ident_stream >> p )
   {
     
      ret = convert_multiface_to_identification_template(p.get_multiface(),
                                                         eye_locs_ptrs,
                                                         &ident_template_size,
                                                         ident_template);
    
      if (ret != 0)
      {
         cerr << "Failed to generate identification template for person "
              << p.get_pid() << ": " << ret << "." << endl;

      } else { 

         ret = identify_template(ident_template, ident_template_size,
                                 num_candidates, candidate_ptrs);

         if (ret != 0)
         {
            cerr << "Bad return value after call to identify_template() for person "
                 << p.get_pid() << endl;

         } else {

            stringstream candidate_file(stringstream::in | stringstream::out);
            candidate_file << candidate_dir << setw(3) <<  setfill('0')
                           << p.get_pid() << ".candidates";
 
            ofstream candidate_stream(candidate_file.str().c_str(), ios::out | ios::binary);

            if ( !candidate_stream )
               fail("Error opening %s for writing.", candidate_file.str().c_str());

            for (size_t rank = 0; rank < num_candidates; rank++)
               candidate_stream << rank << " " << candidates[rank] << endl;

            candidate_stream.close();

         } // end if identify_template failed.

      } // end if template generation failed.

   }

   delete [] ident_template;

   cout << "Searching completed." << endl;

   ident_stream.close();
}

