/*
 * This software was developed at the National Institute of Standards and
 * Technology (NIST) by employees of the Federal Government in the course
 * of their official duties. Pursuant to title 17 Section 105 of the
 * United States Code, this software is not subject to copyright protection
 * and is in the public domain. NIST assumes no responsibility  whatsoever for
 * its use by other parties, and makes no guarantees, expressed or implied,
 * about its quality, reliability, or any other characteristic.
 */

/*
 * Author: George Quinn, NIST
 * Email comments to gw@nist.gov
 */

extern "C"
{
   #define MBE_CLASS_A
   #include "mbe.h"

   #define __STDC_FORMAT_MACROS
   #include <inttypes.h>			// for PRIu32
}

#include "person.h"
#include "mbe_validation.h"

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <time.h>
#include <cstring>	// for memcpy

using namespace std;


const char* score_file    = "./output/verify_1to1/verify_1to1.scr",
          * enroll_file   = "./input/enroll.txt",
          * verif_file    = "./input/search.txt",
          * template_file = "./output/verify_1to1/templates.cat",
          * manifest_file = "./output/verify_1to1/templates.manifest";

const char* img_types[1]    = { "mugshot" };
const uint8_t num_img_types = 1;


class enroll_template
{
   public: 
      enroll_template(const enroll_template& t) : pid(t.pid), size(t.size)
      {
         data = new uint8_t[size];
         memcpy(data, t.data, size);
      }

      enroll_template(const uint32_t p, const uint8_t* d, const uint32_t s) :
         pid(p), size(s)
      {
         data = new uint8_t[s];
         memcpy(data, d, s);
      };

      ~enroll_template()
      {
         delete [] data;
         data = 0;
      }

      uint32_t pid,
               size;

      uint8_t* data;
};


int main(int argc, char* argv[])
{

   if (argc < 2)
      fail("usage:\n%s config_dir", argv[0]);

   const char* config_dir = argv[1];


   ifstream enroll_stream(enroll_file, ios::in | ios::binary);
   if ( !enroll_stream )
      fail("Failed to open enrollment file %s for reading.", enroll_file);

   ofstream template_stream(template_file, ios::out | ios::binary);
   if ( !template_stream )
      fail("Failed to open template file %s for writing.", template_file);

   ofstream manifest_stream(manifest_file, ios::out | ios::binary);
   if ( !manifest_stream )
      fail("Failed to open manifest file %s for writing.", manifest_file);


   cout << "Calling initialize_verification()." << endl;
   
   uint32_t ret = initialize_verification(config_dir, img_types, num_img_types);

   if (ret != 0)
      fail("Bad return value from initialize_verification: %" PRIu32 ".", ret);


   uint32_t max_enroll_size,
            max_verif_size;

   ret = get_max_template_sizes(&max_enroll_size, &max_verif_size);

   if (ret != 0)
      fail("Bad return value from get_max_template_sizes(): %" PRIu32 ".", ret);


   uint8_t* enroll_template_data = new uint8_t[ max_enroll_size * person::max_num_faces ];
   uint32_t enroll_template_size;

   person p;                              // wrapper for MULTIFACE object
   
   vector<enroll_template> enroll_templates;
   
   cout << "Making enrollment templates." << endl;
 
   // Iterate over each person in enrollment file.
   while ( enroll_stream >> p )
   {
      ret = convert_multiface_to_enrollment_template(p.get_multiface(),
                                                     &enroll_template_size,
                                                     enroll_template_data);

      if (ret != 0)
         cerr << "Failed to generate template for person " << p.get_pid()
              << ": " << ret << "." << endl;

      enroll_template tmplate( p.get_pid(), enroll_template_data, enroll_template_size );
      enroll_templates.push_back( tmplate );

      manifest_stream << p.get_pid() << " " << enroll_template_size << " "
                      << template_stream.tellp() << endl;

      // Write the template to the template concatenation file.
      template_stream.write((char*)enroll_template_data, enroll_template_size);

      if (template_stream.bad())
         fail("Failed to write to template concatenation file.");
   }

   // House-keeping.
   enroll_stream.close();
   template_stream.close();
   manifest_stream.close();

   delete [] enroll_template_data;
  
 
   cout << "Enrollment completed." << endl;
   
   ifstream verif_stream(verif_file, ios::in | ios::binary);
   if ( !verif_stream )
      fail("Failed to open %s for reading.", verif_file);

   ofstream score_stream(score_file, ios::out | ios::binary);
   if ( !score_stream )
      fail("Failed to open %s for writing.", score_file);
  
   score_stream.precision(score_precision);
   score_stream.setf(ios::fixed, ios::floatfield); 

 
   uint8_t* verif_template = new uint8_t[ max_verif_size * person::max_num_faces ];
   uint32_t verif_template_size;

   uint8_t quality;

   cout << "Making verification templates and matching." << endl;

   // Iterate over each person in search file.
   while ( verif_stream >> p )
   {
      ret = convert_multiface_to_verification_template(p.get_multiface(),
                                                       &verif_template_size,
                                                       verif_template, &quality);
    
      if (ret != 0)
      {
         cerr << "Failed to generate identification template for person "
              << p.get_pid() << ": " << ret << "." << endl;
      }
            
      vector<enroll_template>::iterator iter;

      for (iter = enroll_templates.begin(); iter != enroll_templates.end(); ++iter)
      {
         double similarity;
                 
         ret = match_templates(verif_template, verif_template_size, iter->data,
                               iter->size, &similarity);

         score_stream << ret << " " << similarity << endl;
                
      }
   }

   cout << "Matching completed." << endl;

   score_stream.close();
   verif_stream.close();

   delete [] verif_template;
}

