/*
 * This software was developed at the National Institute of Standards and
 * Technology (NIST) by employees of the Federal Government in the course
 * of their official duties. Pursuant to title 17 Section 105 of the
 * United States Code, this software is not subject to copyright protection
 * and is in the public domain. NIST assumes no responsibility  whatsoever for
 * its use by other parties, and makes no guarantees, expressed or implied,
 * about its quality, reliability, or any other characteristic.
 */

/*
 * Author: George Quinn
 * Email comments to gw@nist.gov
 */

extern "C"
{
   #define MBE_CLASS_B
   #include "mbe.h"

   #define __STDC_FORMAT_MACROS
   #include <inttypes.h>		// for PRIu32
}

#include "person.h"
#include "mbe_validation.h"

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <time.h>
#include <limits>

using namespace std;


const char* enroll_dir    = "./output/verify_1toN/templates",
          * score_file    = "./output/verify_1toN/verify_1toN.scr",
          * enroll_file   = "./input/enroll.txt",
          * verif_file    = "./input/search.txt",
          * template_file = "./output/verify_1toN/templates.cat",
          * manifest_file = "./output/verify_1toN/templates.manifest";

const char* img_types[1] = { "mugshot" };
const uint8_t num_img_types = 1;

const uint32_t num_persons   = 54,
               num_templates = 54,
               num_images    = 88;


int main(int argc, char* argv[])
{

   if (argc < 2)
      fail("usage:\n%s config_dir", argv[0]);

   const char* config_dir = argv[1];

   ifstream enroll_stream(enroll_file, ios::in | ios::binary);
   if ( !enroll_stream )
      fail("Failed to open enrollment file %s for reading.", enroll_file);


   cout << "Calling initialize_enrollment_session()." << endl;

   int32_t ret = initialize_enrollment_session(config_dir, enroll_dir,
                                               num_persons, num_images,
                                               img_types, num_img_types);

   if (ret != 0)
      fail("Bad return value from initialize_enrollment_session(): %" PRIu32 ".", ret);

 
   uint32_t max_enroll_size,
            max_verif_size;

   ret = get_max_template_sizes(&max_enroll_size, &max_verif_size);

   if (ret != 0)
      fail("Bad return value from get_max_template_sizes(): %" PRIu32 ".", ret);


   ofstream template_stream(template_file, ios::out | ios::binary);
   if ( !template_stream )
      fail("Failed to open template file %s for writing.", template_file);

   ofstream manifest_stream(manifest_file, ios::out | ios::binary);
   if ( !manifest_stream )
      fail("Failed to open manifest file %s for writing.", manifest_file);


   EYEPAIR  eye_locs[person::max_num_faces];
   EYEPAIR* eye_locs_ptrs[person::max_num_faces];

   for (int i = 0; i < person::max_num_faces; ++i)
      eye_locs_ptrs[i] = &eye_locs[i];

   person p;                  // wrapper for MULTIFACE object
   uint32_t template_counter = 0;

   uint8_t* enroll_template = new uint8_t[ max_enroll_size * person::max_num_faces ];
   uint32_t enroll_template_size;


   vector<uint32_t> enroll_ids;
   enroll_ids.reserve(num_persons);
   
   cout << "Enrolling templates." << endl;
 
   // Iterate over each person in enrollment file.
   while ( enroll_stream >> p )
   {
      ret = convert_multiface_to_enrollment_template(p.get_multiface(),
                                                     eye_locs_ptrs,
                                                     &enroll_template_size,
                                                     enroll_template);

      if (ret != 0)
      {
         cerr << "Failed to generate template for person " << p.get_pid()
              << ": " << ret << "." << endl;

      } else {
         enroll_ids.push_back(p.get_pid());

         manifest_stream << p.get_pid() << " " << enroll_template_size
                         << " " << template_stream.tellp() << endl;
         
         // Write the template to the template concatenation file.
         template_stream.write((char*)enroll_template, enroll_template_size);

         if (template_stream.bad())
            fail("Failed to write to the template concatenation file.\n");
      }

      ++template_counter;
   }

   if (template_counter != num_templates)
      fail("Should be %" PRIu32 " templates, but read %" PRIu32 ".", num_templates, template_counter);

   // House-keeping.
   enroll_stream.close();
   template_stream.close();
   manifest_stream.close();

   delete [] enroll_template;
 

   cout << "Calling finalize_enrollment()." << endl;

   ret = finalize_enrollment(enroll_dir, template_file, manifest_file);

   if (ret != 0)
      fail("Bad return value from finalize_enrollment: %" PRIu32, ret);
  


   cout << "Enrollment completed." << endl
        << "Calling initialize_feature_extraction_session()." << endl;

   ret = initialize_feature_extraction_session(config_dir, enroll_dir);

   if (ret != 0)
      fail("Bad return value from initialize_feature_extraction_session: %" PRIu32,
           ret);


   cout << "Calling initialize_identification_session()." << endl;

   ret = initialize_identification_session(config_dir, enroll_dir);

   if (ret != 0)
      fail("Bad return value from initialize_identification_session: %" PRIu32,
           ret);

   
   ifstream verif_stream(verif_file, ios::in | ios::binary);
   if ( !verif_stream )
      fail("Failed to open %s for reading.", verif_file);

   ofstream score_stream(score_file, ios::out | ios::binary);
   if ( !score_stream )
      fail("Failed to pen %s for writing.", score_file);
  
   score_stream.precision(score_precision);
   score_stream.setf(ios::fixed, ios::floatfield);

 
   uint8_t* verif_template = new uint8_t[ max_verif_size * person::max_num_faces ];
   uint32_t verif_template_size;

   
   cout << "Making verification templates and matching." << endl;

   // Iterate over each person in search file.
   while ( verif_stream >> p )
   {

      ret = convert_multiface_to_identification_template(p.get_multiface(),
                                                         eye_locs_ptrs,
                                                         &verif_template_size,
                                                         verif_template);

      if (ret != 0)
      {
         cerr << "Failed to generate identification template for person "
              << p.get_pid() << ": " << ret << "." << endl;
      }


      vector<uint32_t>::iterator pid_iter;

      // Compare to each enrollment template. 
      for (pid_iter = enroll_ids.begin(); pid_iter != enroll_ids.end(); ++pid_iter)
      {
          double similarity = -1;
                 
          ret = verify_template(verif_template, verif_template_size,
                                *pid_iter, &similarity);

          score_stream << ret << " " << similarity << endl;
                
      } // end foreach enrollment template
   }

   cout << "Matching completed." << endl;

   score_stream.close();
   verif_stream.close();

   delete [] verif_template;
}

